// -Bitmaps Demo-
// Demonstrates DirectDraw
// Written by J. Farrell

// This program demonstrates clipping, scaling, and image loading
// with DirectDraw. The image is scaled and shifted around the
// screen until you press Escape, at which point the demo ends.

#define WIN32_LEAN_AND_MEAN

// To compile this program, remember what you need to do to set up
// Visual C++ for DirectX. Make sure the DirectX SDK directories
// for libraries and headers are FIRST on their repective lists
// (under Tools menu, go to "Options..."). Then create a new project
// using this and Script1.rc as the only source files, and remember
// to link the ddraw.lib and dxguid.lib libraries to your project!

// includes
#include <windows.h>
#include <windowsx.h>
#include <ddraw.h>
#include "resource.h"

// constants
#define SIZE_MAXIMUM  120
#define SIZE_MINIMUM   40

#define X_MAXIMUM     640
#define X_MINIMUM       0

#define Y_MAXIMUM     480
#define Y_MINIMUM       0

// macros
#define INIT_DXSTRUCT(dxstruct) { ZeroMemory(&dxstruct, sizeof(dxstruct)); dxstruct.dwSize = sizeof(dxstruct); }
#define KEYSTATE(key) ((GetAsyncKeyState(key) & 0x8000) ? TRUE : FALSE)
#define RGB_16BIT555(r, g, b) ((r<<10) | (g<<5) | (b))
#define RGB_16BIT565(r, g, b) ((r<<11) | (g<<5) | (b))

// G L O B A L S /////////////////////////////////////////////////////
// Windows
HINSTANCE hMainInstance;
HWND hMainWindow;

// DirectDraw
LPDIRECTDRAW7 lpdd7 = NULL;
LPDIRECTDRAWSURFACE7 lpddsPrimary = NULL;
LPDIRECTDRAWSURFACE7 lpddsBack = NULL;
LPDIRECTDRAWSURFACE7 lpddsImage = NULL;
LPDIRECTDRAWCLIPPER  lpddClipper = NULL;
DDSURFACEDESC2 ddsd;

//////////////////////////////////////////////////////////////////////

// prototypes
LRESULT CALLBACK WindowProc(HWND, UINT, WPARAM, LPARAM);
int Initialize(void);
int Load_Bitmap(char*, int, int, LPDIRECTDRAWSURFACE7);
void Shutdown(void);

//////////////////////////////////////////////////////////////////////
// WINMAIN
// Program entry point. Handles the window setup and the contains the
// program's main loop.
//////////////////////////////////////////////////////////////////////
int APIENTRY WinMain(HINSTANCE hinstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	WNDCLASSEX winclass; // our window class
	HWND hwnd;           // handle to a window
	MSG msg;             // a message variable

	int cx = 320, 
		cy = 200;        // coordinates of image
	int nSize = 100;     // half-width of image currently
	int dirSize = 1,     // is size increasing or decreasing?
	    dirX = 2,        // what about x-position?
		dirY = 2;        // and y-position?

	// fill out the winclass structure
	winclass.cbSize =        sizeof(WNDCLASSEX);
	winclass.style =         CS_DBLCLKS | CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	winclass.lpfnWndProc =   WindowProc;        
	winclass.cbClsExtra =    0;                  // extra class info
	winclass.cbWndExtra =    0;                  // extra window info
	winclass.hInstance =     hinstance;	         
	winclass.hIcon =         LoadIcon(NULL, IDI_WINLOGO);
	winclass.hCursor =       LoadCursor(NULL, IDC_ARROW);
	winclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	winclass.lpszMenuName =  NULL;
	winclass.lpszClassName = "TME Class";  
	winclass.hIconSm =       LoadIcon(NULL, IDI_WINLOGO);
	
	// register class
	if (!RegisterClassEx(&winclass))
		return(0);
	
	// create a window
	if (!(hwnd = CreateWindowEx(NULL,            // extended style
			     "TME Class",                    // class identifier
				 "Clipping and Scaling Demo",    // window title
				 WS_POPUP | WS_VISIBLE,          // window style
				 0, 0,							 // initial position
				 640, 480,						 // initial size
				 NULL,                           // handle to parent
				 NULL,                           // handle to menu
				 hinstance,                      // app handle
				 NULL)))                         // extra parameters
		return(0);

	// set globals
	hMainWindow = hwnd;
	hMainInstance = hinstance;

	// initialize, test for success
	if (Initialize())
	{
		// dest RECT for image copy
		RECT dest;

		// main program loop
		while (TRUE)
		{
			// get the time
			DWORD dwStart = GetTickCount();

			// check the message queue
			if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				if (msg.message == WM_QUIT) // exit loop on quit
					break;
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			// update the image size and position
			cx += dirX;
			cy += dirY;
			nSize += dirSize;

			// check for boundaries on x, y, and size
			if (((cx <= X_MINIMUM) && (dirX < 0)) ||
				((cx >= X_MAXIMUM) && (dirX > 0)))
				dirX = -dirX;

			if (((cy <= Y_MINIMUM) && (dirY < 0)) ||
				((cy >= Y_MAXIMUM) && (dirY > 0)))
				dirY = -dirY;

			if (((nSize <= SIZE_MINIMUM) && (dirSize < 0)) ||
				((nSize >= SIZE_MAXIMUM) && (dirSize > 0)))
				dirSize = -dirSize;

			// set the destination rect according to current
			// size and position
			SetRect(&dest, cx-nSize, cy-nSize, cx+nSize, cy+nSize);

			// clear the back buffer with a color fill
			DDBLTFX fx;
			INIT_DXSTRUCT(fx);
			fx.dwFillColor = 0; // black is zero in every color depth
			lpddsBack->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &fx);

			// copy the image to the back buffer
			lpddsBack->Blt(&dest, lpddsImage, NULL, DDBLT_WAIT, NULL);

			// now blit the back buffer to the primary surface
			lpddsPrimary->Blt(NULL, lpddsBack, NULL, DDBLT_WAIT, NULL);

			// Make sure you understand why we're doing this instead
			// of copying the image directly onto the primary surface
			// each time. If we did that, then we'd be color-filling
			// the primary surface on every frame, and so the animation
			// would flicker. We don't want that, so we use the back
			// buffer instead. This is called double-buffering.

			// check escape key
			if (KEYSTATE(VK_ESCAPE))
				PostQuitMessage(0);

			// lock to 30 FPS (33ms per frame)
			while ((GetTickCount() - dwStart) < 33);
		}
	}
	else
		MessageBox(hMainWindow, "Initialization failed!", "Error", MB_OK | MB_ICONEXCLAMATION);
	
	// shut everything down
	Shutdown();

	// give Windows the shutdown code
	return(msg.wParam);
}

LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg,
							WPARAM wparam, LPARAM lparam)
{
    // lazy man's message handler
	return(DefWindowProc(hwnd, msg, wparam, lparam));
}

///////////////////////////////////////////////////////////////
// INITIALIZE
// Creates all the DirectX objects we'll need for this demo,
// sets the screen resolution to 640x480x16.
///////////////////////////////////////////////////////////////
int Initialize(void)
{
	// initialize COM
	if (FAILED(CoInitialize(NULL)))
		return(FALSE);

	// create DirectDraw object
	if (FAILED(CoCreateInstance(CLSID_DirectDraw, NULL, CLSCTX_ALL,
								IID_IDirectDraw7, (void**)&lpdd7)))
		return(FALSE);
	
	// initialize DirectDraw object
	if (FAILED(lpdd7->Initialize(NULL)))
		return(FALSE);

	// set cooperative level to fullscreen
	if (FAILED(lpdd7->SetCooperativeLevel(hMainWindow, DDSCL_FULLSCREEN | 
		                                               DDSCL_EXCLUSIVE |
													   DDSCL_ALLOWREBOOT)))
		return(FALSE);

	// hide the mouse
	ShowCursor(FALSE);

	// set 640x480x16 resolution
	if (FAILED(lpdd7->SetDisplayMode(640, 480, 16, 0, 0)))
		return(FALSE);

	// create our primary surface
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsPrimary, NULL)))
		return(FALSE);

	// create the back buffer
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth = 640;
	ddsd.dwHeight = 480;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsBack, NULL)))
		return(FALSE);

	// create the clipper
	if (FAILED(lpdd7->CreateClipper(0, &lpddClipper, NULL)))
		return(FALSE);

	// create and set the clip list
	LPRGNDATA lpClipList = (LPRGNDATA)malloc(sizeof(RGNDATAHEADER) +
										     sizeof(RECT));
	RECT rcBoundary = {0, 0, 640, 480};
	memcpy(lpClipList->Buffer, &rcBoundary, sizeof(RECT)); // set list
	lpClipList->rdh.dwSize = sizeof(RGNDATAHEADER);        // structure size
	lpClipList->rdh.iType = RDH_RECTANGLES;                // bound type
	lpClipList->rdh.nCount = 1;                            // num of bounds
	lpClipList->rdh.nRgnSize = sizeof(RECT);               // size of Buffer
	lpClipList->rdh.rcBound = rcBoundary;                  // bounding RECT
	if (FAILED(lpddClipper->SetClipList(lpClipList, 0)))
	{
		// always free the mallocs!
		free(lpClipList);
		return(FALSE);
	}

	// attach clipper to back buffer
	if (FAILED(lpddsBack->SetClipper(lpddClipper)))
	{
		free(lpClipList);
		return(FALSE);
	}
	
	// free the memory
	free(lpClipList);

	// create the image surface (image is 200x200)
	INIT_DXSTRUCT(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth = 200;
	ddsd.dwHeight = 200;
	if (FAILED(lpdd7->CreateSurface(&ddsd, &lpddsImage, NULL)))
		return(FALSE);

	// load the image onto the surface
	if (!Load_Bitmap(MAKEINTRESOURCE(IDB_BITMAP1), 0, 0, lpddsImage))
		return(FALSE);

	// all set!
	return(TRUE);
}

//////////////////////////////////////////////////////////////////////
// LOAD_BITMAP
// Loads a bitmap resource onto the specified DirectDraw surface.
//////////////////////////////////////////////////////////////////////
int Load_Bitmap(char *name, int x, int y, LPDIRECTDRAWSURFACE7 dest_surface)
{
	HBITMAP hbitmap = NULL;
	BITMAP bmp;
	HDC image_dc, surface_dc;

	// load bitmap resource
	if (!(hbitmap = (HBITMAP)LoadImage(hMainInstance, name, 
		                               IMAGE_BITMAP, 0, 0, 
					     			   LR_CREATEDIBSECTION)))
		return(FALSE);

	// create device context for image
	if (!(image_dc = CreateCompatibleDC(NULL)))
		return(FALSE);

	// select bitmap into the DC
	if (!(SelectObject(image_dc, hbitmap)))
		return(FALSE);

	// get surface device context
	if (FAILED(dest_surface->GetDC(&surface_dc)))
		return(FALSE);

	// get image dimensions
	GetObject(hbitmap, sizeof(BITMAP), &bmp);
    int dx = bmp.bmWidth;
    int dy = bmp.bmHeight;

	// copy image to surface
	if (!(BitBlt(surface_dc, x, y, dx, dy, image_dc, 0, 0, SRCCOPY)))
		return(FALSE);
	
	// get rid of all that stuff we used
	dest_surface->ReleaseDC(surface_dc);
	DeleteDC(image_dc);
	DeleteObject(hbitmap);

	return(TRUE);
}

///////////////////////////////////////////////////////////////
// SHUTDOWN
// Release the objects we created.
///////////////////////////////////////////////////////////////
void Shutdown(void)
{
	if (lpddsImage)
	{
		// murder the image surface!
		lpddsImage->Release();
		lpddsImage = NULL;
	}

	if (lpddClipper)
	{
		// obliterate the clipper!
		lpddClipper->Release();
		lpddClipper = NULL;
	}

	if (lpddsBack)
	{
		// annihilate the back buffer!
		lpddsBack->Release();
		lpddsBack = NULL;
	}

	if (lpddsPrimary)
	{
		// kill the primary surface!
		lpddsPrimary->Release();
		lpddsPrimary = NULL;
	}

	if (lpdd7)
	{
		// blow away DirectDraw!
		lpdd7->Release();
		lpdd7 = NULL;
	}

	// destroy the window!
	CloseWindow(hMainWindow);

	// assassinate COM! :)
	CoUninitialize();
}
